@@ -1,4 +1,3 @@
-static const char CVSID[] = "$Id: textSel.c,v 1.5 2001/08/14 08:37:16 jlous Exp $";
 /*******************************************************************************
 *									       *
 * textSel.c - Selection and clipboard routines for NEdit text widget		       *
@@ -8,7 +7,8 @@ static const char CVSID[] = "$Id: textSel.c,v 1.5 2001/08/14 08:37:16 jlous Exp
 * This is free software; you can redistribute it and/or modify it under the    *
 * terms of the GNU General Public License as published by the Free Software    *
 * Foundation; either version 2 of the License, or (at your option) any later   *
-* version.							               *
+* version. In addition, you may distribute version of this program linked to   *
+* Motif or Open Motif. See README for details.                                 *
 * 									       *
 * This software is distributed in the hope that it will be useful, but WITHOUT *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        *
@@ -25,28 +25,41 @@ static const char CVSID[] = "$Id: textSel.c,v 1.5 2001/08/14 08:37:16 jlous Exp
 * Written by Mark Edel							       *
 *									       *
 *******************************************************************************/
+
+#ifdef HAVE_CONFIG_H
+#include "../config.h"
+#endif
+
+#include "textSel.h"
+#include "textP.h"
+#include "text.h"
+#include "textDisp.h"
+#include "textBuf.h"
+#include "../util/misc.h"
+#include "../util/nedit_malloc.h"
+
 #include <stdio.h>
 #include <string.h>
 #include <limits.h>
 
 #include <Xm/Xm.h>
 #include <Xm/CutPaste.h>
+#include <Xm/Text.h>
 #include <X11/Xatom.h>
 #if XmVersion >= 1002
 #include <Xm/PrimitiveP.h>
 #endif
 
-#include "textBuf.h"
-#include "textDisp.h"
-#include "textP.h"
-#include "textSel.h"
+#ifdef HAVE_DEBUG_H
+#include "../debug.h"
+#endif
+
 
 #define N_SELECT_TARGETS 7
-#define N_CLIP_TARGETS 4
-#define N_ATOMS 10
+#define N_ATOMS 11
 enum atomIndex {A_TEXT, A_TARGETS, A_MULTIPLE, A_TIMESTAMP,
 	A_INSERT_SELECTION, A_DELETE, A_CLIPBOARD, A_INSERT_INFO,
-	A_ATOM_PAIR, A_MOTIF_DESTINATION};
+	A_ATOM_PAIR, A_MOTIF_DESTINATION, A_COMPOUND_TEXT};
 
 /* Results passed back to the convert proc processing an INSERT_SELECTION
    request, by getInsertSelection when the selection to insert has been
@@ -70,7 +83,7 @@ typedef struct {
 } selectNotifyInfo;
 
 static void modifiedCB(int pos, int nInserted, int nDeleted,
-	int nRestyled, char *deletedText, void *cbArg);
+	int nRestyled, const char *deletedText, void *cbArg);
 static void sendSecondary(Widget w, Time time, Atom sel, int action,
 	char *actionText, int actionTextLen);
 static void getSelectionCB(Widget w, XtPointer clientData, Atom *selType,
@@ -144,7 +157,7 @@ void CopyToClipboard(Widget w, Time time)
     /* Get the selected text, if there's no selection, do nothing */
     text = BufGetSelectionText(((TextWidget)w)->text.textD->buffer);
     if (*text == '\0') {
-    	XtFree(text);
+    	NEditFree(text);
     	return;
     }
 
@@ -153,25 +166,37 @@ void CopyToClipboard(Widget w, Time time)
     length = strlen(text);
     BufUnsubstituteNullChars(text, ((TextWidget)w)->text.textD->buffer);
     
+    /* Shut up LessTif */
+    if (SpinClipboardLock(XtDisplay(w), XtWindow(w)) != ClipboardSuccess) {
+        NEditFree(text);
+        return;
+    }
+    
     /* Use the XmClipboard routines to copy the text to the clipboard.
        If errors occur, just give up.  */
-    stat = XmClipboardStartCopy(XtDisplay(w), XtWindow(w),
-    	    s=XmStringCreateSimple("NEdit"), time, w, NULL, &itemID);
+    s = XmStringCreateSimple("NEdit");
+    stat = SpinClipboardStartCopy(XtDisplay(w), XtWindow(w), s,
+    	    time, w, NULL, &itemID);
     XmStringFree(s);
-    if (stat != ClipboardSuccess)
+    if (stat != ClipboardSuccess) {
+        SpinClipboardUnlock(XtDisplay(w), XtWindow(w));
     	return;
-#ifdef notdef
+    }
+
     /* Note that we were previously passing length + 1 here, but I suspect
        that this was inconsistent with the somewhat ambiguous policy of
        including a terminating null but not mentioning it in the length */
-#endif
-    if (XmClipboardCopy(XtDisplay(w), XtWindow(w), itemID, "STRING",
+
+    if (SpinClipboardCopy(XtDisplay(w), XtWindow(w), itemID, "STRING",
     	    text, length, 0, NULL) != ClipboardSuccess) {
-    	XtFree(text);
+    	NEditFree(text);
+        SpinClipboardEndCopy(XtDisplay(w), XtWindow(w), itemID);
+        SpinClipboardUnlock(XtDisplay(w), XtWindow(w));
     	return;
     }
-    XtFree(text);
-    XmClipboardEndCopy(XtDisplay(w), XtWindow(w), itemID);
+    NEditFree(text);
+    SpinClipboardEndCopy(XtDisplay(w), XtWindow(w), itemID);
+    SpinClipboardUnlock(XtDisplay(w), XtWindow(w));
 }
 
 /*
@@ -251,7 +276,7 @@ void MovePrimarySelection(Widget w, Time time, int isColumnar)
 ** Insert the X CLIPBOARD selection at the cursor position.  If isColumnar,
 ** do an BufInsertCol for a columnar paste instead of BufInsert.
 */
-void InsertClipboard(Widget w, Time time, int isColumnar)
+void InsertClipboard(Widget w, int isColumnar)
 {
     unsigned long length, retLength;
     textDisp *textD = ((TextWidget)w)->text.textD;
@@ -270,13 +295,24 @@ void InsertClipboard(Widget w, Time time, int isColumnar)
        size of the data that be transferred via the clipboard, and are
        generally slower and buggier, they do preserve the clipboard across
        widget destruction and even program termination. */
-    if (XmClipboardInquireLength(XtDisplay(w), XtWindow(w), "STRING", &length)
-    	    != ClipboardSuccess || length == 0)
+    if (SpinClipboardInquireLength(XtDisplay(w), XtWindow(w), "STRING", &length)
+    	    != ClipboardSuccess || length == 0) {
+        /*
+         * Possibly, the clipboard can remain in a locked state after
+         * a failure, so we try to remove the lock, just to be sure.
+         */
+        SpinClipboardUnlock(XtDisplay(w), XtWindow(w));
     	return;
-    string = XtMalloc(length+1);
-    if (XmClipboardRetrieve(XtDisplay(w), XtWindow(w), "STRING", string,
+    }
+    string = (char*)NEditMalloc(length+1);
+    if (SpinClipboardRetrieve(XtDisplay(w), XtWindow(w), "STRING", string,
     	    length, &retLength, &id) != ClipboardSuccess || retLength == 0) {
-    	XtFree(string);
+    	NEditFree(string);
+        /*
+         * Possibly, the clipboard can remain in a locked state after
+         * a failure, so we try to remove the lock, just to be sure.
+         */
+        SpinClipboardUnlock(XtDisplay(w), XtWindow(w));
     	return;
     }
     string[retLength] = '\0';
@@ -285,7 +321,7 @@ void InsertClipboard(Widget w, Time time, int isColumnar)
        else, or give up, warn, and refuse */
     if (!BufSubstituteNullChars(string, retLength, buf)) {
 	fprintf(stderr, "Too much binary data, text not pasted\n");
-	XtFree(string);
+	NEditFree(string);
 	return;
     }
 
@@ -294,7 +330,12 @@ void InsertClipboard(Widget w, Time time, int isColumnar)
     	cursorPos = TextDGetInsertPosition(textD);
     	cursorLineStart = BufStartOfLine(buf, cursorPos);
     	column = BufCountDispChars(buf, cursorLineStart, cursorPos);
-    	BufInsertCol(buf, column, cursorLineStart, string, NULL, NULL);
+        if (((TextWidget)w)->text.overstrike) {
+	    BufOverlayRect(buf, cursorLineStart, column, -1, string, NULL,
+			   NULL);
+	} else {
+	    BufInsertCol(buf, column, cursorLineStart, string, NULL, NULL);
+	}
     	TextDSetInsertPosition(textD,
     	    	BufCountForwardDispChars(buf, cursorLineStart, column));
 	if (((TextWidget)w)->text.autoShowInsertPos)
@@ -302,7 +343,7 @@ void InsertClipboard(Widget w, Time time, int isColumnar)
     } else
     	TextInsertAtCursor(w, string, NULL, True,
 		((TextWidget)w)->text.autoWrapPastedText);
-    XtFree(string);
+    NEditFree(string);
 }
 
 /*
@@ -336,7 +377,7 @@ void TakeMotifDestination(Widget w, Time time)
 ** result, since later callbacks will see the second modifications first).
 */
 static void modifiedCB(int pos, int nInserted, int nDeleted,
-	int nRestyled, char *deletedText, void *cbArg)
+	int nRestyled, const char *deletedText, void *cbArg)
 {
     TextWidget w = (TextWidget)cbArg;
     Time time = XtLastTimestampProcessed(XtDisplay((Widget)w));
@@ -348,13 +389,12 @@ static void modifiedCB(int pos, int nInserted, int nDeleted,
     if ((isOwner && selected) || (!isOwner && !selected))
     	return;
 
-    /* If we own the selection and the selection is now empty, give it up */
-    if (isOwner && !selected) {
-    	XtDisownSelection((Widget)w, XA_PRIMARY, time);
-    	w->text.selectionOwner = False;
-    	return;
-    }
-    
+    /* Don't disown the selection here.  Another application (namely: klipper)
+       may try to take it when it thinks nobody has the selection.  We then
+       lose it, making selection-based macro operations fail.  Disowning
+       is really only for when the widget is destroyed to avoid a convert
+       callback from firing at a bad time. */
+
     /* Take ownership of the selection */
     if (!XtOwnSelection((Widget)w, XA_PRIMARY, time, convertSelectionCB,
     	    loseSelectionCB, NULL))
@@ -399,7 +439,7 @@ static void sendSecondary(Widget w, Time time, Atom sel, int action,
        notify event is never returned */
     XConvertSelection(XtDisplay(w), sel, getAtom(disp, A_INSERT_SELECTION),
     	    getAtom(disp, A_INSERT_INFO), XtWindow(w), time);
-    cbInfo = (selectNotifyInfo *)XtMalloc(sizeof(selectNotifyInfo));
+    cbInfo = (selectNotifyInfo *)NEditMalloc(sizeof(selectNotifyInfo));
     cbInfo->action = action;
     cbInfo->timeStamp = time;
     cbInfo->widget = (Widget)w;
@@ -426,15 +466,14 @@ static void getSelectionCB(Widget w, XtPointer clientData, Atom *selType,
  
     /* Confirm that the returned value is of the correct type */
     if (*type != XA_STRING || *format != 8) {
-    	if (value != NULL)
-    	    XtFree((char *)value);
+        NEditFree(value);
     	return;
     }
     
     /* Copy the string just to make space for the null character (this may
        not be necessary, XLib documentation claims a NULL is already added,
        but the Xt documentation for this routine makes no such claim) */
-    string = XtMalloc(*length + 1);
+    string = (char*)NEditMalloc(*length + 1);
     memcpy(string, (char *)value, *length);
     string[*length] = '\0';
     
@@ -442,8 +481,8 @@ static void getSelectionCB(Widget w, XtPointer clientData, Atom *selType,
        else, or give up, warn, and refuse */
     if (!BufSubstituteNullChars(string, *length, textD->buffer)) {
 	fprintf(stderr, "Too much binary data, giving up\n");
-	XtFree(string);
-	XtFree((char *)value);
+	NEditFree(string);
+	NEditFree(value);
 	return;
     }
     
@@ -458,11 +497,11 @@ static void getSelectionCB(Widget w, XtPointer clientData, Atom *selType,
     } else
     	TextInsertAtCursor(w, string, NULL, False,
 		((TextWidget)w)->text.autoWrapPastedText);
-    XtFree(string);
+    NEditFree(string);
     
     /* The selection requstor is required to free the memory passed
        to it via value */
-    XtFree((char *)value);
+    NEditFree(value);
 }
 
 /*
@@ -481,14 +520,13 @@ static void getInsertSelectionCB(Widget w, XtPointer clientData,Atom *selType,
  
     /* Confirm that the returned value is of the correct type */
     if (*type != XA_STRING || *format != 8 || value == NULL) {
-    	if (value != NULL)
-    	    XtFree((char *)value);
+        NEditFree(value);
     	*resultFlag = UNSUCCESSFUL_INSERT;
     	return;
     }
     
     /* Copy the string just to make space for the null character */
-    string = XtMalloc(*length + 1);
+    string = (char*)NEditMalloc(*length + 1);
     memcpy(string, (char *)value, *length);
     string[*length] = '\0';
     
@@ -496,19 +534,19 @@ static void getInsertSelectionCB(Widget w, XtPointer clientData,Atom *selType,
        else, or give up, warn, and refuse */
     if (!BufSubstituteNullChars(string, *length, buf)) {
 	fprintf(stderr, "Too much binary data, giving up\n");
-	XtFree(string);
-	XtFree((char *)value);
+	NEditFree(string);
+	NEditFree(value);
 	return;
     }
     
     /* Insert it in the text widget */
     TextInsertAtCursor(w, string, NULL, True,
 	    ((TextWidget)w)->text.autoWrapPastedText);
-    XtFree(string);
+    NEditFree(string);
     *resultFlag = SUCCESSFUL_INSERT;
     
     /* This callback is required to free the memory passed to it thru value */
-    XtFree((char *)value);
+    NEditFree(value);
 }
 
 /*
@@ -523,8 +561,7 @@ static void getExchSelCB(Widget w, XtPointer clientData, Atom *selType,
 {
     /* Confirm that there is a value and it is of the correct type */
     if (*length == 0 || value == NULL || *type != XA_STRING || *format != 8) {
-    	if (value != NULL)
-    	    XtFree((char *)value);
+        NEditFree(value);
     	XBell(XtDisplay(w), 0);
     	BufSecondaryUnselect(((TextWidget)w)->text.textD->buffer);
     	return;
@@ -557,8 +594,12 @@ static Boolean convertSelectionCB(Widget w, Atom *selType, Atom *target,
     int getFmt, result = INSERT_WAITING;
     XEvent nextEvent;
     
-    /* target is text or string */
-    if (*target == XA_STRING || *target == getAtom(display, A_TEXT)) {
+    /* target is text, string, or compound text */
+    if (*target == XA_STRING || *target == getAtom(display, A_TEXT) ||
+        *target == getAtom(display, A_COMPOUND_TEXT)) {
+        /* We really don't directly support COMPOUND_TEXT, but recent
+           versions gnome-terminal incorrectly ask for it, even though
+           don't declare that we do.  Just reply in string format. */
     	*type = XA_STRING;
     	*value = (XtPointer)BufGetSelectionText(buf);
     	*length = strlen((char *)*value);
@@ -569,7 +610,7 @@ static Boolean convertSelectionCB(Widget w, Atom *selType, Atom *target,
     
     /* target is "TARGETS", return a list of targets we can handle */
     if (*target == getAtom(display, A_TARGETS)) {
-	targets = (Atom *)XtMalloc(sizeof(Atom) * N_SELECT_TARGETS);
+	targets = (Atom *)NEditMalloc(sizeof(Atom) * N_SELECT_TARGETS);
 	targets[0] = XA_STRING;
 	targets[1] = getAtom(display, A_TEXT);
 	targets[2] = getAtom(display, A_TARGETS);
@@ -631,8 +672,15 @@ static Boolean convertSelectionCB(Widget w, Atom *selType, Atom *target,
 
 static void loseSelectionCB(Widget w, Atom *selType)
 {
-    ((TextWidget)w)->text.selectionOwner = False;
-    BufUnselect(((TextWidget)w)->text.textD->buffer);
+    TextWidget tw = (TextWidget)w;
+    selection *sel = &tw->text.textD->buffer->primary;
+    char zeroWidth = sel->rectangular ? sel->zeroWidth : 0;
+    
+    /* For zero width rect. sel. we give up the selection but keep the 
+        zero width tag. */
+    tw->text.selectionOwner = False;
+    BufUnselect(tw->text.textD->buffer);
+    sel->zeroWidth = zeroWidth;
 }
 
 /*
@@ -682,7 +730,7 @@ static Boolean convertMotifDestCB(Widget w, Atom *selType, Atom *target,
     
     /* target is "TARGETS", return a list of targets it can handle */
     if (*target == getAtom(display, A_TARGETS)) {
-	targets = (Atom *)XtMalloc(sizeof(Atom) * 3);
+	targets = (Atom *)NEditMalloc(sizeof(Atom) * 3);
 	targets[0] = getAtom(display, A_TARGETS);
 	targets[1] = getAtom(display, A_TIMESTAMP);
 	targets[2] = getAtom(display, A_INSERT_SELECTION);
@@ -770,9 +818,8 @@ static void selectNotifyEH(Widget w, XtPointer data, XEvent *event,
     	XBell(XtDisplay(w), 0);
     	BufSecondaryUnselect(buf);
         XtDisownSelection(w, XA_SECONDARY, e->time);
-        if (cbInfo->actionText != NULL)
-            XtFree((char *)cbInfo->actionText);
-    	XtFree((char *)cbInfo);
+        NEditFree(cbInfo->actionText);
+    	NEditFree(cbInfo);
     	return;
     }
 
@@ -782,7 +829,7 @@ static void selectNotifyEH(Widget w, XtPointer data, XEvent *event,
     if (cbInfo->action == REMOVE_SECONDARY) {
     	BufRemoveSecSelect(buf);
     } else if (cbInfo->action == EXCHANGE_SECONDARY) {
-	string = XtMalloc(cbInfo->length + 1);
+	string = (char*)NEditMalloc(cbInfo->length + 1);
 	memcpy(string, cbInfo->actionText, cbInfo->length);
 	string[cbInfo->length] = '\0';
 	selStart = buf->secondary.start;
@@ -799,13 +846,12 @@ static void selectNotifyEH(Widget w, XtPointer data, XEvent *event,
 	    }
 	} else
 	    fprintf(stderr, "Too much binary data\n");
-	XtFree(string);
+	NEditFree(string);
     }
     BufSecondaryUnselect(buf);
     XtDisownSelection(w, XA_SECONDARY, e->time);
-    if (cbInfo->actionText != NULL)
-        XtFree((char *)cbInfo->actionText);
-    XtFree((char *)cbInfo);
+    NEditFree(cbInfo->actionText);
+    NEditFree(cbInfo);
 }
 
 /*
@@ -822,9 +868,8 @@ static void selectNotifyTimerProc(XtPointer clientData, XtIntervalId *id)
     XtRemoveEventHandler(cbInfo->widget, 0, True, selectNotifyEH, cbInfo);
     BufSecondaryUnselect(buf);
     XtDisownSelection(cbInfo->widget, XA_SECONDARY, cbInfo->timeStamp);
-    if (cbInfo->actionText != NULL)
-        XtFree((char *)cbInfo->actionText);
-    XtFree((char *)cbInfo);
+    NEditFree(cbInfo->actionText);
+    NEditFree(cbInfo);
 }
 
 /*
@@ -836,7 +881,7 @@ static Atom getAtom(Display *display, int atomNum)
     static Atom atomList[N_ATOMS] = {0};
     static char *atomNames[N_ATOMS] = {"TEXT", "TARGETS", "MULTIPLE",
     	    "TIMESTAMP", "INSERT_SELECTION", "DELETE", "CLIPBOARD",
-    	    "INSERT_INFO", "ATOM_PAIR", "MOTIF_DESTINATION"};
+    	    "INSERT_INFO", "ATOM_PAIR", "MOTIF_DESTINATION", "COMPOUND_TEXT"};
     
     if (atomList[atomNum] == 0)
     	atomList[atomNum] = XInternAtom(display, atomNames[atomNum], False);
